#include <glib.h>
#include <string.h>
#include <stdio.h>
#include "pdu.h"

static int nibble(char c)
{
	if (c >= '0' && c <= '9')
		return c-'0';
	if (c >= 'A' && c <= 'F')
		return c-'A'+10;
	if (c >= 'a' && c <= 'f')
		return c-'a'+10;
	return 0;
}

static int byte(const char *ptr)
{
	return (nibble(ptr[0]) << 4) | nibble(ptr[1]);
}

static int byte_sw_bcd(const char *ptr)
{
	return nibble(ptr[1])*10+nibble(ptr[0]);
}


void
pdu_dump(struct pdu *pdu)
{
	if (pdu->smsc)
		printf("smsc: type 0x%x number '%s'\n", pdu->smsc->type, pdu->smsc->number);
	printf("pid 0x%x\n", pdu->pid);
	printf("dcs 0x%x\n", pdu->dcs);
	printf("%02d:%02d:%02d (+%02d) %02d.%02d.%04d\n", pdu->hour, pdu->minute, pdu->second, pdu->timezone, pdu->day, pdu->month, pdu->year);
	if (pdu->text) {
		printf("Text:\n");
		printf("%s\n", pdu->text);
	}
}

char *
pdu_text_decode(const char *ptr, int len)
{
	char *ret,*dst;
	int i,bits,x,l,c;

	ret=g_malloc(len*2+1);
	dst=ret;

	l=0;
	bits=0;
	for (i = 0 ; i < len ; i++) {
		if (bits != 8) {
			x=byte(ptr); ptr+=2;
		} else {
			bits=0;
			x=l;
			l=0;
		}
		c=(x << bits) | l;
		l=(x >> (7-bits));
		bits++;
		*dst++=c & 0x7f;
	}
	*dst++='\0';

	return ret;
}

gchar *
pdu_text_encode(char *text)
{
	gchar *ret,*dst;
	int i,bits,len,x,l,c;

	if (! text)
		return g_strdup("");
	len=strlen(text);
	ret=g_malloc(len*2+2);
	dst=ret;

	l=0;
	bits=8;
	for (i = 0 ; i < len ; i++) {
		x=*text++;
		c=(x << (8-bits) & 0xff) | l;
		l=x >> bits;
		if (bits != 8) {
			sprintf(dst,"%02X",c);
			dst+=2;
		} else {
			bits=0;
			l=x;
		}
		bits++;
		*dst+=2;
	}
	sprintf(dst,"%02X",l);
	dst+=2;
	return ret;
}

struct pdu_number *
pdu_number_new(const char *number)
{
	struct pdu_number *n;

	if (! number)
		return NULL;
	n=g_new0(struct pdu_number, 1);
	if (number[0] == '+') {
		n->number=g_strdup(number+1);
		n->type=0x91;
	} else {
		n->number=g_strdup(number);
		n->type=0x81;
	}
	return n;
}

gchar *
pdu_number_encode(struct pdu_number *n)
{
	gchar *ret,*p,*q,o1,o2;
	int len;

	if (!n || ! n->number)
		return g_strdup("");
	len=strlen(n->number);
	ret=g_malloc(len*2+4);
	p=ret;
	sprintf(p,"%02X",n->type);
	p+=2;
	q=n->number;
	while (len > 0) {
		o1=*q++;
		o2=*q++;
		o1-='0';
		if (o2)
			o2-='0';
		else
			o2=15;
		sprintf(p,"%X%X", o2, o1);
		p+=2;
		len-=2;
	}
	return ret;
}

void
pdu_number_free(struct pdu_number *n)
{
	if (n) {
		g_free(n->number);
		g_free(n);
	}
}

struct pdu *
pdu_new(const char *number, const char *smsc, const char *text)
{
	struct pdu *p;
	p=g_new0(struct pdu, 1);
	p->smsc=pdu_number_new(smsc);
	p->flags=0x25;
	p->msg_ref_nr=0;
	p->number=pdu_number_new(number);
	p->pid=0;
	p->dcs=0;
	p->text=g_strdup(text);

	return p;
}

gchar *
pdu_encode(const struct pdu *pdu)
{
	gchar *ret,*smsc,*number,*validity,*text;
	int len=0, textlen=0;

	smsc=pdu_number_encode(pdu->smsc);
	number=pdu_number_encode(pdu->number);
	if (pdu->number && pdu->number->number)
		len=strlen(pdu->number->number);
	validity=g_strdup("");
	text=pdu_text_encode(pdu->text);
	if (pdu->text)
		textlen=strlen(pdu->text);
	ret=g_strdup_printf("%02X%s%02X%02X%02X%s%02X%02X%s%02X%s", strlen(smsc)/2, smsc, pdu->flags, pdu->msg_ref_nr, len, number, pdu->pid, pdu->dcs, validity, textlen, text);
	g_free(text);
	g_free(validity);
	g_free(number);
	g_free(smsc);

	return ret;
}

struct pdu *
pdu_decode(const char *pdu)
{
	const char *ptr=pdu;
	int type,smsc_len,dat,first,len;

	smsc_len=byte(ptr); ptr+=2;
	type=byte(ptr); ptr+=2;

	struct pdu *p;

	p=g_new0(struct pdu, 1);
	p->smsc=NULL;
	p->number=NULL;
	p->text=NULL;

	if (type == 145)
	smsc_len--;
	while (smsc_len > 0) {
		dat=byte(ptr); ptr+=2;	
		dat >>=4;
		smsc_len--;
	}
	first=byte(ptr); ptr+=2;
	smsc_len=byte(ptr); ptr+=2;
	type=byte(ptr); ptr+=2;
	while (smsc_len > 0) {
		dat=byte(ptr); ptr+=2;	
		dat >>=4;
		smsc_len-=2;
	}
	p->pid=byte_sw_bcd(ptr); ptr+=2;
	p->dcs=byte_sw_bcd(ptr); ptr+=2;
	p->year=byte_sw_bcd(ptr)+1900; ptr+=2;
	if (p->year < 1950)
		p->year+=100;
	p->month=byte_sw_bcd(ptr); ptr+=2;
	p->day=byte_sw_bcd(ptr); ptr+=2;
	p->hour=byte_sw_bcd(ptr); ptr+=2;
	p->minute=byte_sw_bcd(ptr); ptr+=2;
	p->second=byte_sw_bcd(ptr); ptr+=2;
	p->timezone=byte_sw_bcd(ptr); ptr+=2;	
	len=byte(ptr); ptr+=2;

	p->text=pdu_text_decode(ptr, len);

	return p;
}

void
pdu_free(struct pdu *p)
{
	pdu_number_free(p->number);
	pdu_number_free(p->smsc);
	g_free(p->text);
	g_free(p);
}
